package dev.pfaff.dyl4j;

import dev.pfaff.glass.OS;
import org.jspecify.annotations.NonNull;

import java.io.IOException;
import java.io.InputStream;
import java.lang.foreign.Arena;
import java.lang.foreign.MemorySegment;
import java.lang.foreign.SymbolLookup;
import java.nio.file.*;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.HexFormat;

public final class BundledLibraryLinker {
	private static final String name = "dyl4j";

	private static final Path CACHE_DIR = Path.of(System.getProperty("user.home")).resolve(switch (OS.HOST) {
		case OS.Windows _ -> "AppData\\Local\\" + name + "\\cache";
		case OS.macOS _ -> "Library/Caches/" + name;
		default -> ".local/cache/" + name;
	});

	static {
		try {
			Files.createDirectories(CACHE_DIR);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	public static SymbolLookup createLookup(MemorySegment ms, Arena arena) throws IOException {
		var bytes = ms.asByteBuffer();
		MessageDigest hasher;
		try {
			hasher = MessageDigest.getInstance("SHA-256");
		} catch (NoSuchAlgorithmException e) {
			throw new UnsupportedOperationException(e);
		}
		hasher.update(bytes);
		bytes.rewind();
		var digest = hasher.digest();
		var digestString = HexFormat.of().formatHex(digest);
		var path = CACHE_DIR.resolve(digestString + ".dylib");
		if (!Files.exists(path)) {
			var tmpPath = Files.createTempFile(CACHE_DIR, digestString, null);
			try (var chan = Files.newByteChannel(tmpPath,
					StandardOpenOption.WRITE,
					StandardOpenOption.TRUNCATE_EXISTING)) {
				while (bytes.hasRemaining()) chan.write(bytes);
			}
			try {
				Files.move(tmpPath, path, StandardCopyOption.ATOMIC_MOVE);
			} catch (FileAlreadyExistsException e) {
				// someone else beat us to it.
			}
		}
		return SymbolLookup.libraryLookup(path, arena);
	}

	public static SymbolLookup createLookup(byte @NonNull [] bytes, Arena arena) throws IOException {
		return createLookup(MemorySegment.ofArray(bytes), arena);
	}

	public static SymbolLookup createLookup(InputStream inputStream, Arena arena) throws IOException {
		return createLookup(inputStream.readAllBytes(), arena);
	}
}
